@@ -27,13 +27,13 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-#include <stdbool.h>
 #include <stdio.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 
+#include "bool.h"
 #include "version.h"
 #include "config.h"
 #include "utils.h"
@@ -55,6 +55,7 @@ struct mode mode = {
 
 	.do_output = true,
 	.output_linenumbers = true,
+	.output_cheaplinenumbers = false,
 	.output_retain_comments = false,
 	.output_file = NULL,
 
@@ -193,6 +194,16 @@ apply_commandline_macros(void)
 	array_setsize(&commandline_macros, 0);
 }
 
+static
+void
+apply_magic_macro(unsigned num, const char *name)
+{
+	struct place p;
+
+	place_setbuiltin(&p, num);
+	macro_define_magic(&p, name);
+}
+
 static
 void
 apply_builtin_macro(unsigned num, const char *name, const char *val)
@@ -209,6 +220,9 @@ apply_builtin_macros(void)
 {
 	unsigned n = 1;
 
+	apply_magic_macro(n++, "__FILE__");
+	apply_magic_macro(n++, "__LINE__");
+
 #ifdef CONFIG_OS
 	apply_builtin_macro(n++, CONFIG_OS, "1");
 #endif
@@ -768,6 +782,7 @@ static const struct flag_option flag_options[] = {
 	{ "MG",                         &mode.depend_assume_generated, true },
 	{ "MP",                         &mode.depend_issue_fakerules,  true },
 	{ "P",                          &mode.output_linenumbers,      false },
+	{ "p",                          &mode.output_cheaplinenumbers, true },
 	{ "Wcomment",                   &warns.nestcomment,    true },
 	{ "Wendif-labels",              &warns.endiflabels,    true },
 	{ "Werror",                     &mode.werror,          true },
@@ -934,7 +949,7 @@ check_arg_option(const char *opt, const struct place *argplace, char *arg)
 	return false;
 }
 
-DEAD static
+DEAD PF(2, 3) static
 void
 usage(const char *progname, const char *fmt, ...)
 {
@@ -946,7 +961,7 @@ usage(const char *progname, const char *fmt, ...)
 	va_end(ap);
 	fprintf(stderr, "\n");
 
-	fprintf(stderr, "Usage: %s [options] [infile [outfile]]\n", progname);
+	fprintf(stderr, "usage: %s [options] [infile [outfile]]\n", progname);
 	fprintf(stderr, "Common options:\n");
 	fprintf(stderr, "   -C               Retain comments\n");
 	fprintf(stderr, "   -Dmacro[=def]    Predefine macro\n");
@@ -1040,7 +1055,7 @@ main(int argc, char *argv[])
 	init();
 
 	for (i=1; i<argc; i++) {
-		if (argv[i][0] != '-') {
+		if (argv[i][0] != '-' || argv[i][1] == 0) {
 			break;
 		}
 		place_setcommandline(&cmdplace, i, 1);
@@ -1065,9 +1080,15 @@ main(int argc, char *argv[])
 	}
 	if (i < argc) {
 		inputfile = argv[i++];
+		if (!strcmp(inputfile, "-")) {
+			inputfile = NULL;
+		}
 	}
 	if (i < argc) {
 		outputfile = argv[i++];
+		if (!strcmp(outputfile, "-")) {
+			outputfile = NULL;
+		}
 	}
 	if (i < argc) {
 		usage(progname, "Extra non-option argument %s", argv[i]);
